#include "mouse_zoom.h"
#include <cstdlib>

template <class T> inline void swap (T &a, T &b)
{
    T tmp;
    tmp = a; 
    a = b;
    b = tmp;
}

MouseZoom::MouseZoom (const int doublebuffer)
{
    dbuff = doublebuffer;
    allocate_map ();
}

MouseZoom::~MouseZoom ()
{
    delete_map ();
    delete [] map;
}

void MouseZoom::delete_map ()
{
    for (int i=0; i<4; i++)
	delete[] map[i];
}    
    
void MouseZoom::allocate_map ()
{
    map = new Int16* [4];
    map[0]  = new Int16 [20];
    map[1] = new Int16 [20];
    map[2] = new Int16 [20];
    map[3] = new Int16 [20];
}

void MouseZoom::rect_read_outline (Screencoord xll, Screencoord yll, 
				   Screencoord xur, Screencoord yur) 
{
    if (xll > xur)
        swap (xll,xur);
    if (yll > yur)
        swap (yll,yur);

    delete_map ();

    map[0] = new Int16 [(xur-xll+4)*3]; 
    map[1] = new Int16 [(yur-yll+4)*3];
    map[2] = new Int16 [(xur-xll+4)*3];
    map[3] = new Int16 [(yur-yll+4)*3];

    rectread (xll-1,yll-1, xur+1, yll+1, map[0]);
    rectread (xur-1,yll-1, xur+1, yur+1, map[1]);
    rectread (xll-1,yur-1, xur+1, yur+1, map[2]);
    rectread (xll-1,yll-1, xll+1, yur+1, map[3]);
}

void MouseZoom::rect_write_outline (Screencoord xll, Screencoord yll, 
				    Screencoord xur, Screencoord yur) 
{
    if (xll > xur)
        swap (xll,xur);
    if (yll > yur)
        swap (yll,yur);

    rectwrite (xll-1,yll-1, xur+1, yll+1, map[0]);
    rectwrite (xur-1,yll-1, xur+1, yur+1, map[1]);
    rectwrite (xll-1,yur-1, xur+1, yur+1, map[2]);
    rectwrite (xll-1,yll-1, xll+1, yur+1, map[3]);
}

int MouseZoom::zoom (Coord *xl, Coord *xr, Coord *yl, Coord*yr,
			   const int zoom, const int ortho)
{
    Int16 mxold=0, myold=0, d_mx, d_my;
    Int32 xs, ys, x0,y0;
    Coord x1,y1, x2,y2, xscale, yscale;

    if (dbuff)
	frontbuffer (1);

    if (zoom == 1)
        rect_write_outline (mx,my, mx1, my1);

    qread (&mx);
    qread (&my);
    getsize (&xs,&ys);
    getorigin (&x0,&y0);

    mx -= x0; my -= y0;

    if (mx > (xs-2))
        mx = xs-2;
    else if (mx < 2)
        mx = 1;
    
    if (my > (ys-2))
        my = ys-2;
    else if (my < 1)
        my = 1;

    xscale = (*xr-*xl)/(xs-1);
    yscale = (*yr-*yl)/(ys-1);
    x1 = *xl + mx*xscale;
    y1 = *yl + my*yscale;
    unqdevice (LEFTMOUSE);
    
    bool first = false;
    xs -=2;
    ys -=2;


    while (getbutton (LEFTMOUSE)){
        mx1 = getvaluator (MOUSEX) - x0 ;
        my1 = getvaluator (MOUSEY) - y0;

        if (ortho) {
            d_mx = mx1-mx; d_my = my1-my;
            if ( abs (d_mx) != abs (d_my)) {
                if (d_my >0)
                    my1 = my + abs (d_mx);
                else
                    my1 = my - abs (d_mx);
                }
        }               

        if (mx1 > xs)
            mx1 = xs;
        else if (mx1 < 2)
            mx1 = 1;

        if (my1 > ys)
            my1 = ys;
        else if (my1 < 1)
            my1 = 1;

        x2 = *xl + mx1*xscale;
        y2 = *yl + my1*yscale;
        
        if (!first) {       
            color (WHITE);
            rect_read_outline (mx,my, mx1, my1);
            rect (x1,y1,x2 ,y2);
            sleep (0);
            mxold = mx1; myold = my1;
            first = true;
        }
        
        if ( (mx1 - mxold) != 0 || (my1-myold) != 0) {
            color (WHITE);
            rect_write_outline (mx,my, mxold, myold);
            rect_read_outline (mx,my, mx1, my1);
            rect (x1,y1, x2,y2);
            sleep (0);
            mxold = mx1; myold = my1;
        }
        usleep (5000);
    }               
    rect_write_outline (mx,my, mxold,myold);
    
    if (x1 > x2) 
        swap (x1, x2);
    if (y1 > y2) 
        swap (y1, y2);
                    
    rect_read_outline (mx,my, mx1, my1);
    qdevice (LEFTMOUSE);

    if (x1 != x2 && y1 != y2) {
        rect (x1,y1,x2, y2);
        sleep (0);
	if (dbuff)
	    backbuffer (1);
	*xl = x1;
	*xr = x2;
	*yl = y1;
	*yr = y2;
        return 1;
    }
    else {
	if (dbuff)
	    backbuffer (1);
        return 0;
    }
}
